# IMPORTANT: Ensure same with slxpy matlab package

model_config_name = "model.toml"
env_config_name = "env.toml"
metadata_version = "1.0.0"
metadata_name = "metadata.json"
model_dir = "model"
project_ir_name = "ir.json"
